#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#ifndef _MSC_VER
#include "config.h"
#endif
#include "WebSrv.h"

#ifdef DEBUG
#define debug printf
#else
static int debug(char *fmt, ...) { return 0; }
#endif

static char *memstr(char *a, int s, char *b)
{
	int l;
	char *p;
	
	l=strlen(b);
	for(p=a;p<a+s-l;p++)
	{
		if(!memcmp(p,b,l))
			return(p);
	}
	return(NULL);
}
#ifdef _MSC_VER
#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#else
#ifndef HAVE_STRNCASECMP
static int strncasecmp(char *a, char*b, int n)
{
	char *c,*d,*p;
	int o;
	
	c=malloc(n+1);
	memcpy(c,a,n);
	c[n]='\0';
	for(p=c;*p;p++)
		if(*p>='a'&&*p<='a')
			*p-='a'-'A';
	d=strdup(b);
	for(p=d;*p;p++)
		if(*p>='a'&&*p<='a')
			*p-='a'-'A';
	o=strcmp(c,d);
	free(c);
	free(d);
	return(o);
}
#endif
#endif

WebSrv *WebSrv_Create()
{
	WebSrv *ws;
	
	ws=(WebSrv*)malloc(sizeof(WebSrv));
	memset(ws,0,sizeof(WebSrv));
	return(ws);
}

TCPsocket WebSrv_Open(WebSrv *websrv, Uint16 port)
{
	IPaddress ip;
	SDLNet_ResolveHost(&ip, NULL, port);
	websrv->sock=SDLNet_TCP_Open(&ip);
	if(websrv->sock)
	{
		websrv->set=SDLNet_AllocSocketSet(1);
		SDLNet_TCP_AddSocket(websrv->set, websrv->sock);
	}
	return(websrv->sock);
}

int WebSrv_Poll(WebSrv *websrv, Uint32 timeout)
{
	return(SDLNet_CheckSockets(websrv->set,timeout));
}

char *WebSrv_ClientReadLine(WebSrvClient *client)
{
	static char *buf=NULL;
	int len=0, err;
	char c;
	
	if(buf)
		free(buf);
	buf=NULL;
	while((err=SDLNet_TCP_Recv(client->sock,&c,1))==1)
	{
		if(c=='\r')
			continue;
		buf=(char*)realloc(buf,len+1);
		buf[len]=c;
		if(c=='\n')
		{
			buf[len]='\0';
			return(buf);
		}
		len++;
	}
	if(err<0)
	{
		if(buf)
			free(buf);
		buf=NULL;
	}
	return(buf);
}

void WebSrv_ParseParams(WebSrvURL *url, const char *buf)
{
	const char *p,*ee;
	const char *e=buf-1;
	do
	{
		url->params=(WebSrvParams*)realloc(url->params,sizeof(WebSrvParams)*(url->num_params+1));
		memset(&url->params[url->num_params],0,sizeof(WebSrvParams));
		p=e+1;
		e=strchr(p,'&');
		if(e)
			*(char*)e=0;
		ee=strchr(p,'=');
		if(ee)
			*(char*)ee=0;
		url->params[url->num_params].name=WebSrv_dehex(p);
		if(ee)
		{
			url->params[url->num_params].data=WebSrv_dehex(ee+1);
			url->params[url->num_params].size=strlen(url->params[url->num_params].data);
			*(char*)ee='=';
		}
		url->num_params++;
		if(e)
			*(char*)e='&';
	} while(e);
}

void WebSrv_ParseMultipart(WebSrvClient *c)
{
	char *p,*e,*ee;
	int blen;

	blen=strlen(c->boundary);
	//fwrite(c->post,c->post_length,1,stdout); printf("\n");
	if(strncmp(c->post,"--",2) || strncmp(c->boundary, c->post+2, blen))
	{
		debug("boundary not on first line of Multipart POST!\n");
		return;
	}
	for(p=c->post+2+blen;p;p=e)
	{
		while(*p=='\n' || *p=='\r')
			p++;
		e=p;
		e=memstr(e,c->post_length+(char*)c->post-e,c->boundary);
		if(!e)
		{
			debug("end part boundary not found!\n");
			return;
		}
		e[-4]='\0';

		p=strstr(p,"Content-Disposition: form-data; name=\"");
		if(p)
		{
			p+=38;
			ee=strchr(p,'\"');
			if(ee)
			{
				*ee='\0';
				c->url.params=(WebSrvParams*)realloc(c->url.params,sizeof(WebSrvParams)*(c->url.num_params+1));
				memset(&c->url.params[c->url.num_params],0,sizeof(WebSrvParams));
				c->url.params[c->url.num_params].name=strdup(p);
				c->url.params[c->url.num_params].filename=NULL;
				*ee='\"';
				// filename
				ee=strstr(p,"; filename=\"");
				if(ee)
				{
					p=ee+12;
					ee=strchr(p,'\n');
					if(ee)
					{
						char* eol = ee;
						*eol ='\0';
						ee = strrchr(p,'\"');
						*eol ='\n';
					}
					if(ee)
					{
						*ee='\0';
						c->url.params[c->url.num_params].filename=strdup(p);
						*ee='\"';
					}
				}
				// data
				p=strstr(p,"\r\n\r\n");
				if(p)
				{
					p+=4;
					//debug("+++begin+++\n%s\n---end---\n",p);
					c->url.params[c->url.num_params].size=e-p-4;
					if(c->url.params[c->url.num_params].size)
					{
						c->url.params[c->url.num_params].data=malloc(c->url.params[c->url.num_params].size);
						memcpy(c->url.params[c->url.num_params].data,p,c->url.params[c->url.num_params].size);
					}
				}
				c->url.num_params++;
			}
		}
		
		e[-4]='\r';
		e+=blen;
		if(*e=='-')
			break;
	}
}

int WebSrv_ParseURL(WebSrvURL *url)
{
	char *p,*e,*ee;

	p=url->url;
	if(!strncmp(p,"http:",5))
		p+=5;
	e=strchr(p,'/');
	if(e!=p)
	{
		*e=0;
		ee=strchr(p,':');
		if(ee)
			*ee=0;
		url->host=strdup(p);
		if(ee)
		{
			url->port=strtol(ee+1,NULL,0)&0xFFFF;
			*ee=':';
		}
		*e='/';
	}
	p=e;
	e=strchr(p,'?');
	if(e)
		*e=0;
	ee=strrchr(p,'/');
	if(!ee)
		return(-1); /* this meanes it couldn't find the trailing slash...impossible! */
	*ee=0;
	url->path=strdup(p);
	*ee='/';
	url->file=strdup(ee+1);
	if(e)
	{
		*e='?';
		WebSrv_ParseParams(url,e+1);
	}
	return(0);
}

WebSrvClient *WebSrv_Accept(WebSrv *websrv)
{
	WebSrvClient *c;
	TCPsocket sock;
	char *buf,*p;
	int ispost=0, gotlen=0;
	
	sock=SDLNet_TCP_Accept(websrv->sock);
	if(!sock)
		return(NULL);
	c=(WebSrvClient*)malloc(sizeof(WebSrvClient));
	memset(c,0,sizeof(WebSrvClient));
	/*memset(&c->url,0,sizeof(WebSrvURL));*/
	websrv->clients=(WebSrvClient**)realloc(websrv->clients,sizeof(WebSrvClient*)*(websrv->num_clients+1));
	websrv->clients[websrv->num_clients]=c;
	websrv->num_clients++;
	c->sock=sock;
	c->url.encoding=WEBSRV_INVALID;
	while((buf=WebSrv_ClientReadLine(c)) && strlen(buf))
	{
		/*debug("line='%s'\n",buf);*/
		if(!strncasecmp(buf,"GET ",4))
		{
			p=strchr(buf+4,' ');
			if(p)
				*p='\0';
			c->url.url=strdup(buf+4);
			/*debug("url=%s\n",c->url.url);*/
			WebSrv_ParseURL(&c->url);
			c->url.encoding=WEBSRV_URL;
		}
		else if(!strncasecmp(buf,"POST ",5))
		{
			ispost=1;
			p=strchr(buf+5,' ');
			if(p)
				*p='\0';
			c->url.url=strdup(buf+5);
			WebSrv_ParseURL(&c->url);
		}
		else if(!strncasecmp(buf,"Content-length: ",16))
		{
			gotlen=1;
			c->post_length=strtol(buf+16,NULL,0);
		}
		else if(!strncasecmp(buf,"Content-type: ",14))
		{
			if(!strncasecmp(buf+14,"application/x-www-form-urlencoded",33))
				c->url.encoding=WEBSRV_URL;
			if(!strncasecmp(buf+14,"multipart/form-data",19))
			{
				char *s,*e;
				if((s=strstr(buf+14+19,"boundary=")))
				{
					c->url.encoding=WEBSRV_MULTIPART;
					e=strpbrk(s+9," \n\r");
					if(e)
						*e='\0';
					c->boundary=strdup(s+9);
				}
			}
		}
		else
			debug("discarded: %s\n",buf);
	}
	if(!buf)
	{
		WebSrv_CloseClient(websrv,c);
		c=NULL;
	}
	if(ispost && c && c->url.encoding!=WEBSRV_INVALID)
	{
		//debug("post: encoding=%d length=%d boundary=\"%s\"\n",c->url.encoding, c->post_length,c->boundary);
		if(!gotlen)
		{
			debug("POST with no Content-Length!\n");
			WebSrv_SendHTTPResponse(c,400);
			WebSrv_CloseClient(websrv,c);
			c=NULL;
		}
		else if(c->post_length)
		{
			c->post=(unsigned char*)malloc(c->post_length+1);
			if(!c->post)
			{
				debug("POST Content-length too big for memory!\n");
				WebSrv_SendHTTPResponse(c,500); /* out of memory */
				WebSrv_CloseClient(websrv,c);
				c=NULL;
			}
			else
			{
				memset(c->post,0,c->post_length+1);
				gotlen=0;
				while(gotlen!=c->post_length)
				{
					int len;
					if((len=SDLNet_TCP_Recv(c->sock,c->post+gotlen,c->post_length-gotlen))<0)
					{
						debug("POST Content-length not received in full!\n");
						c->post_length=gotlen;
					}
					else
						gotlen+=len;
				}
				if(c->url.encoding==WEBSRV_URL)
					WebSrv_ParseParams(&c->url,c->post);
				else if(c->url.encoding==WEBSRV_MULTIPART && c->boundary)
					WebSrv_ParseMultipart(c);
			}
		}
	}
	return(c);
}

char *WebSrv_enhex(const char *str)
{
	const char *p;
	char *s,*out;

	out=(char*)malloc(strlen(str)*3+1);
    for(p=str,s=out;*p;p++,s++)
    {
		if(*p==' ')
			*s='+';
		else if(!strchr("0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_-.,",*p))
        {
            sprintf(s,"%%%02X",*(unsigned char*)p);
            s+=2;
        }
        else
            *s=*p;
    }
    *s='\0';
    return(out);
}

char *WebSrv_dehex(const char *str)
{
	char *out,*s,*p,hex[3];

	out=s=(char*)malloc(strlen(str)+1);
	hex[2]=0;
    for(p=str;*p;p++,s++)
    {
        if(*p=='%')
        {
            hex[0]=p[1];
            hex[1]=p[2];
            *s=(char)strtol(hex,NULL,16);
			p+=2;
        }
		else if(*p=='+')
			*s=' ';
		else
			*s=*p;
    }
	*s=0;
	out=realloc(out,strlen(out)+1);
    return(out);
}

int WebSrv_ClientSendVString(WebSrvClient *client, const char *format, ...)
{
	char str[4096];
	va_list ap;

	va_start(ap, format);
	vsprintf (str, format, ap);
	va_end(ap);
	str[4095]=0;
	return(WebSrv_ClientSendString(client, str));
}

int WebSrv_SendHTTPResponse(WebSrvClient *client, Uint32 code)
{
	switch(code)
	{
		/* 1XX are not in the standard, they are informational only */
		/* 2XX */
		case 200:
			WebSrv_ClientSendString(client,"HTTP/1.0 200 OK\r\n");
			break;
		case 201:
			WebSrv_ClientSendString(client,"HTTP/1.0 201 Created\r\n");
			break;
		case 202:
			WebSrv_ClientSendString(client,"HTTP/1.0 202 Accepted\r\n");
			break;
		case 204:
			WebSrv_ClientSendString(client,"HTTP/1.0 204 No Content\r\n");
			break;
		/* 3XX */
		case 300:
			WebSrv_ClientSendString(client,"HTTP/1.0 300 Multiple Choices\r\n");
			break;
		case 301:
			WebSrv_ClientSendString(client,"HTTP/1.0 301 Moved Permanently\r\n");
			break;
		case 302:
			WebSrv_ClientSendString(client,"HTTP/1.0 302 Moved Temporarily\r\n");
			break;
		case 304:
			WebSrv_ClientSendString(client,"HTTP/1.0 304 Not Modified\r\n");
			break;
		/* 4XX */
		case 400:
			WebSrv_ClientSendString(client,"HTTP/1.0 400 Bad Request\r\n");
			break;
		case 401:
			WebSrv_ClientSendString(client,"HTTP/1.0 401 Unauthorized\r\n");
			break;
		case 403:
			WebSrv_ClientSendString(client,"HTTP/1.0 403 Forbidden\r\n");
			break;
		case 404:
			WebSrv_ClientSendString(client,"HTTP/1.0 404 Not Found\r\n");
			break;
		/* 5XX */
		case 500:
			WebSrv_ClientSendString(client,"HTTP/1.0 500 Internal Server Error\r\n");
			break;
		case 501:
			WebSrv_ClientSendString(client,"HTTP/1.0 501 Not Implemented\r\n");
			break;
		case 503:
			WebSrv_ClientSendString(client,"HTTP/1.0 503 Service Unavailable\r\n");
			break;
		default:
			return(-1);
			break;
	}
	return(0);
}

void WebSrv_FreeURL(WebSrvURL *url)
{
	int i;

	if(url->url)
		free(url->url);
	if(url->host)
		free(url->host);
	if(url->path)
		free(url->path);
	if(url->file)
		free(url->file);
	if(url->params)
	{
		for(i=0;i<url->num_params;i++)
		{
			if(url->params[i].name)
				free(url->params[i].name);
			if(url->params[i].filename)
				free(url->params[i].filename);
			if(url->params[i].data) /* yes, this can be NULL, Virginia! */
				free(url->params[i].data);
		}
		free(url->params);
	}
	memset(url,0,sizeof(WebSrvURL));
}

void WebSrv_CloseClient(WebSrv *websrv, WebSrvClient *client)
{
	int i;

	if(!client)
		return;
	for(i=0; i<websrv->num_clients && client!=websrv->clients[i]; i++);
	if(!i<websrv->num_clients)
		return;
	SDLNet_TCP_Close(client->sock);
	WebSrv_FreeURL(&client->url);
	if(client->post)
		free(client->post);
	if(client->boundary)
		free(client->boundary);
	free(client);
	memmove(&websrv->clients[i], &websrv->clients[i+1], sizeof(WebSrvClient*)*(websrv->num_clients-i-1));
	websrv->num_clients--;
	if(websrv->num_clients-i)
		websrv->clients=(WebSrvClient**)realloc(websrv->clients,sizeof(WebSrvClient*)*(websrv->num_clients-i));
	else
	{
		if(websrv->clients)
			free(websrv->clients);
		websrv->clients=NULL;
	}
}

void WebSrv_Close(WebSrv *websrv)
{
	SDLNet_TCP_Close(websrv->sock);
	websrv->sock=NULL;
	while(websrv->num_clients)
		WebSrv_CloseClient(websrv, websrv->clients[0]);
	websrv->num_clients=0;
	SDLNet_FreeSocketSet(websrv->set);
	websrv->set=NULL;
}

void WebSrv_Destroy(WebSrv *websrv)
{
	WebSrv_Close(websrv);
	free(websrv);
}
